from urllib.parse import urljoin

import requests

from .constants import *


class Api(object):
    def __init__(self, header_options, base_url=DEFAULT_SUPERGOOD_BASE_URL):
        self.base_url = base_url
        self.header_options = header_options
        self.event_sink_url = None
        self.error_sink_url = None
        self.config_pull_url = None

    def set_logger(self, logger):
        self.log = logger

    # Remote config fetching
    def set_config_pull_url(self, endpoint):
        self.config_pull_url = urljoin(self.base_url, endpoint)

    def get_config(self):
        if not self.config_pull_url:
            raise Exception(ERRORS["UNINITIALIZED"])
        response = requests.get(self.config_pull_url, headers=self.header_options)
        if response.status_code == 401:
            raise Exception(ERRORS["UNAUTHORIZED"])
        elif response.status_code != 200:
            raise Exception(ERRORS["FETCHING_CONFIG"])
        return response.json()

    # Event posting
    def set_event_sink_url(self, endpoint):
        self.event_sink_url = urljoin(self.base_url, endpoint)

    def post_events(self, payload):
        if not self.event_sink_url:
            raise Exception(ERRORS["UNINITIALIZED"])
        response = requests.post(
            self.event_sink_url, json=payload, headers=self.header_options
        )
        if response.status_code == 401:
            raise Exception(ERRORS["UNAUTHORIZED"])
        if response.status_code != 200 and response.status_code != 201:
            raise Exception(ERRORS["POSTING_EVENTS"])
        return response.json()

    # Error posting
    def set_error_sink_url(self, endpoint):
        self.error_sink_url = urljoin(self.base_url, endpoint)

    def post_errors(self, data, exc_info, message):
        if not self.error_sink_url:
            raise Exception(ERRORS["UNINITIALIZED"])
        json = {"payload": data, "error": str(exc_info), "message": message}
        try:
            response = requests.post(
                self.error_sink_url, json=json, headers=self.header_options
            )
            return response.status_code
        except Exception:
            self.log.warning(f"Failed to report error to {self.error_sink_url}")



import http.client
from uuid import uuid4

from ..constants import REQUEST_ID_KEY

HTTPS_PORT = http.client.HTTPS_PORT


def patch(cache_request, cache_response):
    _original_read = http.client.HTTPResponse.read
    _original_getresponse = http.client.HTTPConnection.getresponse
    _original_request = http.client.HTTPConnection.request
    _original_getheaders = http.client.HTTPResponse.getheaders

    def _wrap_read(httpResponse, amt=None):
        response_object = httpResponse
        response_body = _original_read(httpResponse, amt)
        request_id = getattr(response_object, REQUEST_ID_KEY)
        response_headers = _original_getheaders(httpResponse)
        response_status = response_object.status
        response_status_text = response_object.reason
        cache_response(
            request_id=request_id,
            response_body=response_body,
            response_headers=response_headers,
            response_status=response_status,
            response_status_text=response_status_text,
        )
        return response_body

    def _wrap_getresponse(httpConnection):
        response_object = _original_getresponse(httpConnection)
        try:
            request_id = getattr(httpConnection, REQUEST_ID_KEY)
            setattr(response_object, REQUEST_ID_KEY, request_id)
        except AttributeError:
            pass
        return response_object

    def _wrap_request(
        httpConnection,
        method,
        path,
        body=None,
        headers={},
        encode_chunked=False,
        **kwargs,
    ):
        request_id = str(uuid4())
        setattr(httpConnection, REQUEST_ID_KEY, request_id)
        scheme = "https" if httpConnection.port == HTTPS_PORT else "http"
        url = f"{scheme}://{httpConnection.host}{path}"
        cache_request(request_id, url, method, body, headers)
        return _original_request(
            httpConnection,
            method,
            path,
            body=body,
            headers=headers,
            encode_chunked=encode_chunked,
            **kwargs,
        )

    http.client.HTTPResponse.read = _wrap_read
    http.client.HTTPConnection.getresponse = _wrap_getresponse
    http.client.HTTPConnection.request = _wrap_request



import signal

SIGNALS = [
    signal.SIGTERM,
    signal.SIGINT,
]
REQUEST_ID_KEY = "_supergood_request_id"
GZIP_START_BYTES = b"\x1f\x8b"
DEFAULT_SUPERGOOD_BYTE_LIMIT = 500000
DEFAULT_SUPERGOOD_CONFIG_URL = "https://api.supergood.ai/config/"
DEFAULT_SUPERGOOD_BASE_URL = "https://api.supergood.ai/"
DEFAULT_SUPERGOOD_CONFIG = {
    "flushInterval": 1000,
    "configInterval": 10000,
    "eventSinkEndpoint": "/events",
    "errorSinkEndpoint": "/errors",
    "remoteConfigEndpoint": "/config",
    "ignoredDomains": [],
    "ignoreRedaction": False,
    "logRequestHeaders": True,
    "logRequestBody": True,
    "logResponseHeaders": True,
    "logResponseBody": True,
}

ERRORS = {
    "CACHING_RESPONSE": "Error Caching Response",
    "CACHING_REQUEST": "Error Caching Request",
    "DUMPING_DATA_TO_DISK": "Error Dumping Data to Disk",
    "POSTING_EVENTS": "Error Posting Events",
    "POSTING_ERRORS": "Error Posting Errors",
    "FETCHING_CONFIG": "Error Fetching Config",
    "WRITING_TO_DISK": "Error writing to disk",
    "TEST_ERROR": "Test Error for Testing Purposes",
    "UNINITIALIZED": "Client not properly initialized",
    "UNAUTHORIZED": "Unauthorized: Invalid Client ID or Secret. Exiting.",
    "NO_CLIENT_ID": "No Client ID Provided, set SUPERGOOD_CLIENT_ID or pass it as an argument",
    "NO_CLIENT_SECRET": "No Client Secret Provided, set SUPERGOOD_CLIENT_SECRET or pass it as an argument",
    "UNKNOWN": "Client received unexpected value",
    "REDACTION": "Client failed to redact sensitive keys",
    "LOCK_STATE": "Client lock state ambiguous",
}



from threading import Timer


class RepeatingThread(object):
    """
    A function to help run a thread every `interval` seconds
    `skip_first_interval` will not wait for the first interval before running
    """

    def __init__(self, func, interval, skip_first_interval=False, *args, **kwargs):
        self._thread = None
        self.interval = interval
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self._running = False
        self.skip_first_interval = skip_first_interval

    def _run_one(self):
        self._running = False
        self.start()
        self.func(*self.args, **self.kwargs)

    def start(self):
        if not self._running:
            self._thread = Timer(self.interval, self._run_one)
            self._thread.daemon = True
            self._thread.start()
            self._running = True

    def cancel(self):
        if self._thread:
            self._thread.cancel()
        self._running = False



import json
import re
from dataclasses import dataclass
from typing import Dict, List, Tuple, Union
from urllib.parse import urlparse

import tldextract


@dataclass
class SensitiveKey:
    """
    Key-level config
    key_path: path of key to action on
    action: 'REDACT' (redacts value) ('HASH' and 'IGNORE' not supported for now)
    """

    key_path: str
    action: str


@dataclass
class EndpointConfiguration:
    """
    API-level config
    regex: Regex used to uniquely identify the endpoint
    location: Where to find the value to test the regex against
    action: 'Allow' (no-ops), 'Ignore' (does not cache)
    sensitive_keys: Keys to redact from the request and response
    """

    endpoint_id: str
    regex: re.Pattern
    location: str
    action: str
    sensitive_keys: List[SensitiveKey]


@dataclass
class VendorConfiguration:
    """
    Vendor-level config
    id: vendor UUID
    endpoints: List of known endpoints
    """

    domain: str
    vendor_id: str
    endpoints: Dict[str, EndpointConfiguration]


def get_endpoint_test_val(
    location,
    url=None,
    request_body=None,
    request_headers=None,
):
    """
    Uses the location to find the correct value to check endpoint regex against
    """
    # Only uses provided TLD cache. Doesnt make call for updates
    extract = tldextract.TLDExtract(suffix_list_urls=())
    if location == "path":
        return urlparse(url).path
    elif location == "url":
        return url
    elif location == "domain":
        return extract(url).domain
    elif location == "subdomain":
        return extract(url).subdomain
    elif location == "requestHeaders":
        return json.dumps(request_headers)
    elif location == "requestBody":
        return json.dumps(request_body)
    else:
        return ""


def get_vendor_endpoint_from_config(
    remote_config,
    url=None,
    request_body=None,
    request_headers=None,
) -> Tuple[Union[None, VendorConfiguration], Union[None, EndpointConfiguration]]:
    """
    Using the url, request_body, and request_headers
    matches to the vendors/endpoints in remote_config
    and returns a tuple of (VendorConfiguration, EndpointConfiguration)
    if it finds a match, otherwise (None, None)
    """
    extract = tldextract.TLDExtract(suffix_list_urls=())
    url_extract = extract(url)
    search = url_extract.fqdn or url_extract.domain
    vendor_config = next(
        (vcfg for vcfg in remote_config.values() if vcfg.domain in search), None
    )
    if vendor_config:
        return vendor_config, next(
            (
                ep
                for ep in vendor_config.endpoints.values()
                if ep.regex.search(
                    get_endpoint_test_val(
                        location=ep.location,
                        url=url,
                        request_body=request_body,
                        request_headers=request_headers,
                    )
                )
            ),
            None,
        )
    else:
        return (None, None)


def parse_remote_config_json(
    config: List[Dict],
) -> Dict[str, VendorConfiguration]:
    remote_config = {}
    for entry in config:
        vendor_id = entry.get("id")
        endpoints = []
        for endpoint in entry.get("endpoints"):
            matchingRegex = endpoint.get("matchingRegex")
            endpointConfiguration = endpoint.get("endpointConfiguration")
            if not endpointConfiguration:
                # Assume 'Allow' and no sensitive keys when conf is empty
                action = "Allow"
                sensitive_keys = []
            else:
                action = endpointConfiguration.get("action")
                sensitive_keys = list(
                    map(
                        lambda key: SensitiveKey(key.get("keyPath"), key.get("action")),
                        endpointConfiguration.get("sensitiveKeys"),
                    )
                )

            regex = re.compile(matchingRegex.get("regex"))
            endpoints.append(
                EndpointConfiguration(
                    endpoint.get("id"),
                    regex,
                    matchingRegex.get("location"),
                    action,
                    sensitive_keys,
                )
            )
        vendor_config = VendorConfiguration(
            vendor_id=vendor_id,
            domain=entry.get("domain"),
            endpoints={ep.endpoint_id: ep for ep in endpoints},
        )
        remote_config[vendor_id] = vendor_config

    return remote_config



import re

import requests

from supergood.api import Api
from supergood.remote_config import parse_remote_config_json
from tests.helper import get_remote_config


class TestRemoteConfig:
    def test_remote_config_parse(self):
        remote_config = get_remote_config()
        parsed_config = parse_remote_config_json(remote_config)
        assert len(parsed_config) == 1
        assert parsed_config["vendor-id"] is not None
        assert len(parsed_config["vendor-id"].endpoints) == 1
        endpoint_config = parsed_config["vendor-id"].endpoints["endpoint-id"]
        assert endpoint_config.regex == re.compile("200")
        assert endpoint_config.location == "path"
        assert endpoint_config.action == "Allow"
        assert endpoint_config.sensitive_keys == []

    def test_client_ignores_before_config(self, httpserver, supergood_client):
        # Not a perfect way of simulating it but good enough
        supergood_client.remote_config = None  # config not pulled
        httpserver.expect_request("/200").respond_with_json({"key": "val"})
        requests.get(httpserver.url_for("/200"))
        assert supergood_client._request_cache == {}
        assert supergood_client._response_cache == {}
        supergood_client._get_config()  # Now there's a config
        httpserver.expect_request("/200").respond_with_json({"key": "val"})
        requests.get(httpserver.url_for("/200"))
        supergood_client.flush_cache()
        args = Api.post_events.call_args[0][0]
        assert len(args) == 1
        assert args[0]["response"]["body"] == {"key": "val"}



from uuid import uuid4

import httpx

from ..constants import REQUEST_ID_KEY


def patch(cache_request, cache_response):
    _original_handle_request = httpx.HTTPTransport.handle_request
    _original_response_read = httpx.Response.read

    def _wrap_handle_request(
        httpTransport: httpx.HTTPTransport, request: httpx.Request
    ):
        request_id = str(uuid4())
        cache_request(
            request_id,
            str(request.url),
            request.method,
            b"".join(request.stream).decode("utf-8"),
            request.headers,
        )
        response = _original_handle_request(httpTransport, request)
        setattr(response, REQUEST_ID_KEY, request_id)
        return response

    def _wrap_response_read(response: httpx.Response):
        response_body = _original_response_read(response)
        request_id = getattr(response, REQUEST_ID_KEY)
        status_text = response.extensions.get("reason_phrase", None)
        if status_text:
            status_text = status_text.decode("utf-8")
        cache_response(
            request_id,
            response.text,
            response.headers,
            response.status_code,
            status_text,
        )
        return response_body

    httpx.HTTPTransport.handle_request = _wrap_handle_request
    httpx.Response.read = _wrap_response_read



import gzip
import hashlib
import json
import sys
import traceback
from base64 import b64encode
from typing import Tuple

from pydash import get, set_

from .constants import ERRORS, GZIP_START_BYTES
from .remote_config import get_vendor_endpoint_from_config


def hash_value(input):
    _input = input
    hash = hashlib.md5()
    if not input:
        return ""
    encoded = json.dumps(_input, sort_keys=True).encode()
    hash.update(encoded)
    return b64encode(hash.digest()).decode("utf-8")


def get_with_exists(obj, key) -> Tuple[any, bool]:
    """
    obj: a dictionary object, usually a request/response
    key: A keypath to test, in the form key1.key2[0].key3

    If the key path is valid, returns
    (value, True)
    otherwise returns
    (None, False)
    """
    dummy = object()
    test = get(obj, key, dummy)
    # If we get the default value back, the keypath does not exist
    if test != dummy:
        return test, True
    else:
        return None, False


def recursive_size(obj, seen=set(), include_overhead=False):
    """
    NB: Assumes it's operating on a JSON response. Not for general use
    By default does NOT include the overhead of dict overhead, just the sizes of keys and values
    """
    obj_id = id(obj)
    if obj_id in seen:
        return 0
    seen.add(obj_id)
    size = 0
    if isinstance(obj, dict):
        if include_overhead:
            size += sys.getsizeof(obj)
        size += sum([recursive_size(v, seen, include_overhead) for v in obj.values()])
        size += sum([recursive_size(k, seen, include_overhead) for k in obj.keys()])
    else:
        size = sys.getsizeof(obj)
    return size


def describe_data(data):
    """
    data: a portion of a JSON response
    returns: a tuple describing the data: (type, length)
    e.g. describe_data("kazooie") => ("string", 7)

    for integers and floats uses the string length
    for dictionaries attempts to recursively determine the size (not including dictionary overhead)
    """
    typ = type(data).__name__
    if typ == "NoneType":
        return ("null", 0)
    elif typ == "bool":
        return ("boolean", 1)
    elif typ == "str":
        return ("string", len(data))
    elif typ == "int":
        return ("integer", len(str(data)))
    elif typ == "float":
        return ("float", len(str(data)))
    elif typ == "list":
        return ("array", len(data))
    elif typ == "dict":
        return ("object", recursive_size(data))
    else:
        # Expecting only JSON response types here. fail otherwise
        raise Exception(ERRORS["UNKNOWN"])


def deep_redact_(input, keypath, action):
    """
    input: a request or response json blob to be redacted
    keypath: a keypath operating on an array, i.e. containing []
    action: the action to take on that values at that keypath

    To redact a key that is present in one or many response array objects, e.g.
    `response.body.added[].payment_instruments[].number`
    you have to iterate over both (all) arrays to collect all the keys
    that need to be redacted.

    This function redacts (removes) all `keypath` keys in-place on `input`
    and returns metadata about each extracted key, which Supergood uses for schema
    anomaly detection.
    """
    all_keys = []
    for key_segment in keypath.split("."):
        if key_segment.endswith("[]"):
            key_name = key_segment[0:-2]
            new_keys = []
            for keypart in all_keys:
                arr = get(input, ".".join([keypart, key_name]))
                if arr is None:
                    # Key does not exist
                    continue
                for i in range(len(arr)):
                    new_keys.append(".".join([keypart, (key_name + f"[{i}]")]))
            if not new_keys:
                # If at any step we run out of keys, this keypath must not exist anywhere.
                #  we can safely redact nothing and return
                return []
            all_keys = new_keys
        else:
            if len(all_keys) == 0:
                all_keys.append(key_segment)
            else:
                all_keys = list(map(lambda k: ".".join([k, key_segment]), all_keys))

    # Invariant: he first two elements of any keypath will always be
    #  of the form '{request|response}.{body|headers}'
    #  however for standardization we need to transform that to a single key
    #  '{request|response}{Body|Headers}'
    metadata = []
    for key in all_keys:
        keysplit = key.split(".")
        if len(keysplit) == 2:  # to handle top-level redactions like responseBody
            actual_key = ".".join([(keysplit[0] + keysplit[1].capitalize())])
        else:
            actual_key = ".".join(
                [(keysplit[0] + keysplit[1].capitalize())] + keysplit[2:]
            )
        entry = get(input, key)
        (data_type, data_length) = describe_data(entry)
        # NB: The UI only supports `redact` for now, so the clients only support is as well
        set_(input, key, None)
        metadata.append(
            {
                "keyPath": actual_key,
                "type": data_type,
                "length": data_length,
            }
        )

    return metadata


def redact_values(
    input_array,
    remote_config,
    ignore_redaction=False,
):
    """
    input_array: a dictionary, representing a `request, reponse` pair
    remote_config: the SG remote config

    data is redacted in-place
    redaction info is placed in the metadata
    a list is returned indicating indices that should be removed
    """
    remove_indices = []
    if ignore_redaction:
        # add metadata if it doesn't exist, then return
        for i in range(len(input_array)):
            if "metadata" not in input_array[i]:
                input_array[i].update({"metadata": {}})
        return
    for index, data in enumerate(input_array):
        skeys = []
        endpoint = None
        if (
            ("metadata" in data)
            and (vendor_id := data["metadata"].get("vendorId", None))
            and (endpoint_id := data["metadata"].get("endpointId", None))
        ):
            endpoint = remote_config[vendor_id].endpoints[endpoint_id]
        else:
            _, endpoint = get_vendor_endpoint_from_config(
                remote_config,
                url=data["request"].get("url"),
                request_body=data["request"]["body"],
                request_headers=data["request"]["headers"],
            )
        if endpoint:
            # Matched endpoint. Check ignore and then sensitive keys
            if endpoint.action.lower() == "ignore":
                remove_indices.append(index)
                break
            for key in endpoint.sensitive_keys:
                if key.key_path.startswith("response") and not data.get("response"):
                    continue
                key_split = key.key_path.split(".")
                top_level_array = key_split[0].endswith("[]")
                if key_split[0].startswith("requestBody"):
                    keypart1 = "request"
                    keypart2 = "body"
                elif key_split[0].startswith("requestHeaders"):
                    keypart1 = "request"
                    keypart2 = "headers"
                elif key_split[0].startswith("responseBody"):
                    keypart1 = "response"
                    keypart2 = "body"
                elif key_split[0].startswith("responseHeaders"):
                    keypart1 = "response"
                    keypart2 = "headers"
                else:
                    raise Exception(f"unknown keypath {key.key_path}")

                if top_level_array:
                    keypart2 += "[]"
                keypath = ".".join([keypart1, keypart2])
                if len(key_split) > 1:
                    keypath = ".".join([keypath] + key_split[1:])

                if "[]" in keypath:
                    # Keys within an array require iterative redaction
                    redactions = deep_redact_(data, keypath, key.action.lower())
                    skeys += redactions
                else:
                    item, exists = get_with_exists(data, keypath)
                    if not exists:
                        # Key not present, move on
                        continue
                    (data_type, data_length) = describe_data(item)
                    skeys.append(
                        {
                            "keyPath": key.key_path,
                            "type": data_type,
                            "length": data_length,
                        }
                    )
                    # NB: the only action supported for now is `redact`
                    set_(data, keypath, None)
        else:
            # No endpoint, potentially add metadata and move on
            if "metadata" not in data:
                data["metadata"] = {}
        if skeys:
            data["metadata"].update({"sensitiveKeys": skeys})
    return remove_indices


def safe_parse_json(input: str):
    if not input:
        return ""
    try:
        return json.loads(input)
    except Exception:
        return safe_decode(input)


def safe_decode(input, encoding="utf-8"):
    try:
        if isinstance(input, bytes) and input[:2] == GZIP_START_BYTES:
            return gzip.decompress(input)

        if isinstance(input, str):
            return input

        return input.decode(encoding)
    except Exception:
        return str(input)



import pytest
import requests

from supergood.api import Api
from tests.helper import get_config

TEST_BED_URL = "http://supergood-testbed.herokuapp.com"
EXTERNAL_URL = "https://api.ipify.org/?format=json"


class TestIgnoreDomain:
    @pytest.mark.parametrize(
        "supergood_client",
        [
            {
                "config": get_config(
                    ignored_domains=["supergood-testbed.herokuapp.com"]
                ),
            }
        ],
        indirect=True,
    )
    def test_ignores_requests_to_ignored_domains(self, supergood_client):
        requests.get(f"{TEST_BED_URL}/200")
        requests.get(f"{EXTERNAL_URL}")
        supergood_client.flush_cache()
        supergood_client.kill()
        args = Api.post_events.call_args[0][0]
        assert len(args) == 1
        assert EXTERNAL_URL == args[0]["request"]["url"]



import http

import urllib3

from ..constants import REQUEST_ID_KEY


def patch(cache_request, cache_response):
    _original_read_chunked = urllib3.HTTPResponse.read_chunked
    _original_getheaders = http.client.HTTPResponse.getheaders

    def _wrap_read_chunked(urllib3HttpResponse, amt=None, decode_content=None):
        response_object = urllib3HttpResponse._original_response
        response_bytes = []

        for line in _original_read_chunked(urllib3HttpResponse, amt, decode_content):
            response_bytes.append(line)
            yield line

        request_id = getattr(response_object, REQUEST_ID_KEY)
        response_headers = _original_getheaders(response_object)
        response_body = b"".join(response_bytes)
        cache_response(
            request_id=request_id,
            response_body=response_body,
            response_headers=response_headers,
            response_status=response_object.status,
            response_status_text=response_object.reason,
        )

    urllib3.HTTPResponse.read_chunked = _wrap_read_chunked



import asyncio
import json
import re
import sys
import time
import unittest
from urllib.parse import urlparse

import aiohttp
import pytest
import requests
import urllib3
from dotenv import load_dotenv
from pytest_httpserver import HTTPServer
from werkzeug.wrappers import Response

from supergood import Client
from supergood.api import Api
from supergood.constants import ERRORS
from tests.constants import *
from tests.helper import get_remote_config

load_dotenv()

TEST_BED_URL = "http://supergood-testbed.herokuapp.com"


class TestCore:
    def test_captures_all_outgoing_200_http_requests(
        self, httpserver: HTTPServer, supergood_client
    ):
        CALL_COUNT = 5

        for i in range(CALL_COUNT):
            response = requests.get(httpserver.url_for("/200"))

        supergood_client.flush_cache()
        assert Api.post_events.call_args is not None
        args = Api.post_events.call_args[0][0]
        assert args[0]["request"] is not None
        assert args[0]["response"] is not None
        assert len(args) == CALL_COUNT
        supergood_client.kill()

    def test_captures_non_success_status_and_errors(
        self, httpserver: HTTPServer, supergood_client
    ):
        http_error_codes = [400, 401, 403, 404, 500, 501, 502, 503, 504]

        for code in http_error_codes:
            httpserver.expect_request(f"/{code}").respond_with_data(status=code)
            response = requests.get(httpserver.url_for(f"/{code}"))

        supergood_client.flush_cache()

        args = Api.post_events.call_args[0][0]
        assert len(args) == len(http_error_codes)

        responded_codes = set(map(lambda x: x["response"]["status"], args))
        assert responded_codes == set(http_error_codes)
        supergood_client.kill()

    def test_post_requests(self, httpserver: HTTPServer, supergood_client):
        httpserver.expect_request("/200", method="POST").respond_with_json(
            {"payload": "value"}
        )
        response = requests.post(httpserver.url_for("/200"))
        response.raise_for_status()
        json = response.json()
        supergood_client.flush_cache()

        assert json["payload"] == "value"
        args = Api.post_events.call_args[0][0]
        assert len(args) == 1
        supergood_client.kill()

    def test_post_request_for_large_payload(
        self, httpserver: HTTPServer, supergood_client
    ):
        response = requests.post(COMPASS_URL, headers=HEADERS, json=PAYLOAD)
        response.raise_for_status()
        json = response.json()

        supergood_client.flush_cache()

        assert json.get("mapResults", {}).get("totalItems", 0) >= 0
        args = Api.post_events.call_args[0][0][0]
        assert args["request"] is not None
        assert args["response"] is not None
        supergood_client.kill()

    def test_hanging_response(self, httpserver: HTTPServer, supergood_client):
        HANG_TIME_IN_SECONDS = 2
        requests.get(f"{TEST_BED_URL}/200?sleep={HANG_TIME_IN_SECONDS}")
        supergood_client.flush_cache()

        args = Api.post_events.call_args[0][0]
        assert len(args) == 1
        supergood_client.kill()

    def test_posting_errors(self, supergood_client, session_mocker):
        _mock = session_mocker.patch(
            "supergood.api.Api.post_events", side_effect=Exception("Test Error")
        )
        _mock.start()

        requests.get(f"{TEST_BED_URL}/200")

        supergood_client.flush_cache()
        error_message = Api.post_errors.call_args[0][2]
        assert error_message == ERRORS["POSTING_EVENTS"]

        _mock.resetmock()
        session_mocker.patch("supergood.api.Api.post_events").start()
        supergood_client.kill()

    def test_different_http_library(self, httpserver: HTTPServer, supergood_client):
        http = urllib3.PoolManager()
        http.request("GET", f"{TEST_BED_URL}/200")

        supergood_client.flush_cache()

        args = Api.post_events.call_args[0][0]
        assert len(args) == 1
        supergood_client.kill()

    def test_aiohttp_library(self, httpserver: HTTPServer, supergood_client):
        async def aiohttp_get_request():
            async with aiohttp.ClientSession() as session:
                async with session.get(
                    f"{TEST_BED_URL}/200?success=true", headers=HEADERS, json=PAYLOAD
                ) as response:
                    return await response.json()

        async def aiohttp_post_request():
            async with aiohttp.ClientSession() as session:
                async with session.post(
                    f"{TEST_BED_URL}/200?success=true", headers=HEADERS, json=PAYLOAD
                ) as response:
                    return await response.json()

        get_response = asyncio.run(aiohttp_get_request())
        post_response = asyncio.run(aiohttp_post_request())

        supergood_client.flush_cache()

        args = Api.post_events.call_args[0][0]
        assert len(args) == 2
        assert get_response["success"] == "true"
        assert post_response["success"] == "true"
        assert args[0]["response"] is not None
        assert args[0]["request"] is not None
        supergood_client.kill()



import os
from logging import INFO, basicConfig, getLogger

from dotenv import load_dotenv

load_dotenv()
basicConfig(level=INFO)


class Logger:
    def __init__(self, logger_name, config, api):
        self.config = config
        self.log = getLogger(logger_name)
        if os.getenv("SUPERGOOD_LOG_LEVEL") == "debug":
            self.log.setLevel(10)
        self.api = api

    def error(self, error, exc_info, data={}):
        self.log.error(error)
        self.log.error(exc_info)
        self.api.post_errors(data, exc_info, error)

    def info(self, info):
        self.log.info(info)

    def warning(self, warning):
        self.log.warning(warning)

    def debug(self, debug):
        self.log.debug(debug)



#!/usr/bin/env python3

import atexit
import os
import traceback
from base64 import b64encode
from datetime import datetime
from importlib.metadata import version
from threading import Lock, Thread
from urllib.parse import urlparse

from dotenv import load_dotenv

from .api import Api
from .constants import *
from .helpers import redact_values, safe_decode, safe_parse_json
from .logger import Logger
from .remote_config import get_vendor_endpoint_from_config, parse_remote_config_json
from .repeating_thread import RepeatingThread
from .vendors.aiohttp import patch as patch_aiohttp
from .vendors.http import patch as patch_http
from .vendors.httpx import patch as patch_httpx
from .vendors.requests import patch as patch_requests
from .vendors.urllib3 import patch as patch_urllib3

load_dotenv()


class Client(object):
    def __init__(
        self,
        client_id=os.getenv("SUPERGOOD_CLIENT_ID"),
        client_secret_id=os.getenv("SUPERGOOD_CLIENT_SECRET"),
        base_url=os.getenv("SUPERGOOD_BASE_URL"),
        config={},
        metadata={},
    ):
        self.base_url = base_url if base_url else DEFAULT_SUPERGOOD_BASE_URL

        # By default will spin up threads to handle flushing and config fetching
        #  set the appropriate environment variable to override this behavior.
        #  Primarily used during testing/debugging. Most `.env` files wont include these
        auto_flush = True
        auto_config = True
        if os.getenv("SG_OVERRIDE_AUTO_FLUSH"):
            auto_flush = False
        if os.getenv("SG_OVERRIDE_AUTO_CONFIG"):
            auto_config = False

        if "serviceName" not in metadata and "SERVICE_NAME" in os.environ:
            metadata["serviceName"] = os.getenv("SERVICE_NAME")
        self.metadata = metadata

        authorization = f"{client_id}:{client_secret_id}"

        header_options = {
            "Accept": "application/json, text/plain, */*",
            "Content-Type": "application/json",
            "Authorization": "Basic "
            + b64encode(bytes(authorization, "utf-8")).decode("utf-8"),
            "supergood-api": "supergood-py",
            "supergood-api-version": version("supergood"),
        }
        self.base_config = DEFAULT_SUPERGOOD_CONFIG
        self.base_config.update(config)

        self.api = Api(header_options, self.base_url)
        self.api.set_event_sink_url(self.base_config["eventSinkEndpoint"])
        self.api.set_error_sink_url(self.base_config["errorSinkEndpoint"])
        self.api.set_config_pull_url(self.base_config["remoteConfigEndpoint"])
        self.log = Logger(self.__class__.__name__, self.base_config, self.api)

        self.remote_config = None
        if auto_config:
            self.remote_config_initial_pull = Thread(
                daemon=True, target=self._get_config
            )
            self.remote_config_initial_pull.start()
        else:
            self.log.debug("auto config off. Remember to request manually")

        self.remote_config_refresh_thread = RepeatingThread(
            self._get_config, self.base_config["configInterval"] / 1000
        )
        if auto_config:
            self.remote_config_refresh_thread.start()

        self.api.set_logger(self.log)

        self._request_cache = {}
        self._response_cache = {}

        # Initialize patches here
        patch_requests(self._cache_request, self._cache_response)
        patch_urllib3(self._cache_request, self._cache_response)
        patch_http(self._cache_request, self._cache_response)
        patch_aiohttp(self._cache_request, self._cache_response)
        patch_httpx(self._cache_request, self._cache_response)

        self.flush_thread = RepeatingThread(
            self.flush_cache, self.base_config["flushInterval"] / 1000
        )
        self.flush_lock = Lock()
        if auto_flush:
            self.flush_thread.start()
        else:
            self.log.debug("auto flush off, remember to flush manually")

        # On clean exit, or terminated exit - exit gracefully
        atexit.register(self.close)

    def _build_log_payload(self, urls=None, size=None, num_events=None):
        payload = {}
        payload["config"] = self.base_config
        payload["metadata"] = self.metadata
        if urls:
            if len(urls) == 1:
                payload["metadata"].update({"requestUrl": urls[0]})
            else:
                payload["metadata"].update({"requestUrls": urls})
        if size:
            payload["metadata"].update({"payloadSize": size})
        if num_events:
            payload["numberOfEvents"] = num_events
        return payload

    def _should_ignore(
        self,
        host_domain,
        metadata,
        url=None,
        request_body=None,
        request_headers=None,
    ):
        supergood_base_url = urlparse(self.base_url).hostname
        # If we haven't fetched the remote config yet, always ignore
        if (
            self.remote_config is None
            or host_domain == supergood_base_url
            or host_domain in self.base_config["ignoredDomains"]
        ):
            return True

        vendor, endpoint = get_vendor_endpoint_from_config(
            self.remote_config,
            url=url,
            request_body=request_body,
            request_headers=request_headers,
        )
        if endpoint:
            # add endpoint and vendor to metadata for quicker redaction later
            metadata["endpointId"] = endpoint.endpoint_id
            metadata["vendorId"] = vendor.vendor_id
            if endpoint.action.lower() == "ignore":
                return True
        return False

    def _cache_request(self, request_id, url, method, body, headers):
        request = {}
        try:
            host_domain = urlparse(url).hostname
            request["metadata"] = {}
            # Check that we should cache the request
            if not self._should_ignore(
                host_domain,
                request["metadata"],  # we store endpoint id in metadata
                url=url,
                request_body=body,
                request_headers=headers,
            ):
                now = datetime.now().isoformat()
                parsed_url = urlparse(url)
                filtered_body = (
                    ""
                    if not self.base_config["logRequestBody"]
                    else safe_parse_json(safe_decode(body))
                )
                filtered_headers = (
                    {}
                    if (not self.base_config["logRequestHeaders"] or headers is None)
                    else dict(headers)
                )
                request["request"] = {
                    "id": request_id,
                    "method": method,
                    "url": url,
                    "body": filtered_body,
                    "headers": filtered_headers,
                    "path": parsed_url.path,
                    "search": parsed_url.query,
                    "requestedAt": now,
                }
                self._request_cache[request_id] = request
        except Exception:
            payload = self._build_log_payload(
                urls=[url],
            )
            trace = "".join(traceback.format_exc())
            self.log.error(ERRORS["CACHING_REQUEST"], trace, payload)

    def _cache_response(
        self,
        request_id,
        response_body,
        response_headers,
        response_status,
        response_status_text,
    ) -> None:
        request, response = {}, {}
        try:
            # Ignored domains are not in the request cache, so this yields None
            request = self._request_cache.pop(request_id, None)
            if request:
                body = safe_parse_json(safe_decode(response_body))
                filtered_body = "" if not self.base_config["logResponseBody"] else body
                filtered_headers = (
                    {}
                    if not self.base_config["logResponseHeaders"]
                    else dict(response_headers)
                )
                response = {
                    "body": filtered_body,
                    "headers": filtered_headers,
                    "status": response_status,
                    "statusText": response_status_text,
                    "respondedAt": datetime.now().isoformat(),
                }
                self._response_cache[request_id] = {
                    "request": request["request"],
                    "response": response,
                    "metadata": request.get("metadata", {}),
                }
        except Exception:
            url = None
            if request and request.get("request", None):
                url = request.get("request").get("url")
            payload = self._build_log_payload(urls=[url] if url else [])
            trace = "".join(traceback.format_exc())
            self.log.error(ERRORS["CACHING_RESPONSE"], trace, payload)

    def close(self) -> None:
        self.log.debug("Closing client auto-flush, force flushing remaining cache")
        self.flush_thread.cancel()
        self.remote_config_refresh_thread.cancel()
        self.flush_cache(force=True)

    def kill(self) -> None:
        self.log.debug("Killing client auto-flush, deleting remaining cache.")
        self.flush_thread.cancel()
        self.remote_config_refresh_thread.cancel()
        self._request_cache.clear()
        self._response_cache.clear()

    def _get_config(self) -> None:
        try:
            raw_config = self.api.get_config()
            self.remote_config = parse_remote_config_json(raw_config)
        except Exception:
            if self.remote_config:
                self.log.warning("Failed to update remote config")
            else:
                payload = self._build_log_payload()
                trace = "".join(traceback.format_exc())
                self.log.error(ERRORS["FETCHING_CONFIG"], trace, payload)

    def _take_lock(self, block=False) -> bool:
        return self.flush_lock.acquire(blocking=block)

    def _release_lock(self):
        try:
            self.flush_lock.release()
        except RuntimeError:  # releasing a non-held lock
            payload = self._build_log_payload()
            trace = "".join(traceback.format_exc())
            self.log.error(ERRORS["LOCK_STATE"], trace, payload)

    def flush_cache(self, force=False) -> None:
        # if we're not force flushing, and another flush is in progress, just skip
        # if we are force flushing, block until the previous flush completes.
        if self.remote_config is None:
            self.log.info("Config not loaded yet, cannot flush")
            return
        acquired = self._take_lock(block=force)
        if acquired == False:
            self.log.info("Flush already in progress, skipping")
            return
        # FLUSH LOCK PROTECTION START
        response_keys = []
        request_keys = []
        try:
            response_keys = list(self._response_cache.keys())
            request_keys = list(self._request_cache.keys())
            # If there are no responses in cache, just exit
            if len(response_keys) == 0 and not force:
                return

            # If we're forcing a flush but there's nothing in the cache, exit here
            if force and len(response_keys) == 0 and len(request_keys) == 0:
                return

            data = list(self._response_cache.values())

            if force:
                data += list(self._request_cache.values())
            try:
                to_delete = redact_values(
                    data,
                    self.remote_config,
                    self.base_config["ignoreRedaction"],
                )
                if to_delete:
                    data = [
                        item for (ind, item) in enumerate(data) if ind not in to_delete
                    ]
            except Exception:
                urls = []
                for entry in data:
                    if entry.get("request", None):
                        urls.append(entry.get("request").get("url"))
                payload = self._build_log_payload(num_events=len(data), urls=urls)
                trace = "".join(traceback.format_exc())
                self.log.error(ERRORS["REDACTION"], trace, payload)
            else:  # Only post if no exceptions
                self.log.debug(f"Flushing {len(data)} items")
                self.api.post_events(data)
        except Exception:
            payload = self._build_log_payload()
            trace = "".join(traceback.format_exc())
            self.log.error(ERRORS["POSTING_EVENTS"], trace, payload)
        finally:  # always occurs, even from internal returns
            for response_key in response_keys:
                self._response_cache.pop(response_key, None)
            if force:
                for request_key in request_keys:
                    self._request_cache.pop(request_key, None)
            self.flush_lock.release()
            # FLUSH LOCK PROTECTION END



import time

from supergood.repeating_thread import RepeatingThread


class TestRepeatingThread:
    def test_repeating_thread(self):
        class A:
            def inc(self):
                self.i = self.i + 1

            def __init__(self):
                self.i = 0
                self.thread = RepeatingThread(self.inc, 1)

            def start(self):
                self.thread.start()

            def cancel(self):
                self.thread.cancel()

        a = A()
        assert a.thread._thread == None
        a.start()
        for i in range(1, 3):
            time.sleep(1.1)
            assert a.i == i
        a.cancel()
        assert a.thread._thread.finished.is_set()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="supergood",
    version="1.1.1",
    author="Alex Klarfeld",
    description="The Python client for Supergood",
    long_description=long_description,
    long_description_content_type="text/markdown",
    package=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
    py_modules=["supergood"],
    package_dir={"": "src"},
    install_requires=[
        "aiohttp",
        "httpx",
        "jsonpickle",
        "pydash==7.0.1",
        "python-dotenv==1.0.0",
        "requests",
        "tldextract>=5",
        "urllib3>=1.26,<2.0",
    ],
    extras_require={
        "test": [
            "requests",
            "urllib3>=1.26,<2.0",
            "pytest==7.2.1",
            "pytest_httpserver==1.0.8",
            "python-dotenv==1.0.0",
            "Werkzeug",
            "jsonpickle==3.0.1",
            "pytest-mock==3.10.0",
        ]
    },
)
